import configparser
import pathlib


def setup(app):
    app.add_object_type(
        "confval",
        "confval",
        objname="configuration value",
        indextemplate="pair: %s; configuration value",
    )


def get_version():
    # Get current library version without requiring the library to be
    # installed, like ``pkg_resources.get_distribution(...).version`` requires.
    cp = configparser.ConfigParser()
    cp.read(pathlib.Path(__file__).parent.parent / "setup.cfg")
    return cp["metadata"]["version"]


project = "Mopidy-dLeyna"
copyright = "2015-2022 Thomas Kemmer"
version = get_version()
release = version

exclude_patterns = ["_build"]
master_doc = "index"
html_theme = "default"

latex_documents = [
    (
        "index",
        "Mopidy-dLeyna.tex",
        "Mopidy-dLeyna Documentation",
        "Thomas Kemmer",
        "manual",
    )
]

man_pages = [
    (
        "index",
        "mopidy-dleyna",
        "Mopidy-dLeyna Documentation",
        ["Thomas Kemmer"],
        1,
    )
]



from unittest import mock

import pytest


@pytest.fixture
def audio():
    return mock.Mock()


@pytest.fixture
def config():
    return {
        "dleyna": {
            "upnp_browse_limit": 1000,
            "upnp_lookup_limit": 50,
            "upnp_search_limit": 100,
        }
    }


@pytest.fixture
def client():
    client_mock = mock.Mock()
    return client_mock


@pytest.fixture
def backend(config, audio, client):
    from mopidy import backend
    from mopidy_dleyna import library, playback

    backend_mock = mock.Mock(spec=backend.Backend)
    backend_mock.client = client
    backend_mock.library = library.dLeynaLibraryProvider(backend_mock, config)
    backend_mock.playback = playback.dLeynaPlaybackProvider(audio, backend_mock)
    return backend_mock



import collections
import itertools
import logging
import operator

import uritools

from mopidy import backend, models

from . import Extension, translator

logger = logging.getLogger(__name__)

BROWSE_FILTER = ["DisplayName", "Type", "TypeEx", "URI"]

IMAGES_FILTER = ["AlbumArtURL", "URI"]

LOOKUP_FILTER = SEARCH_FILTER = [
    "Album",
    "AlbumArtURL",
    "Artist",
    "Artists",
    "Bitrate",
    "Date",
    "DisplayName",
    "Duration",
    "Genre",
    "TrackNumber",
    "Type",
    "TypeEx",
    "URI",
]

BROWSE_ORDER = {
    models.Ref.ALBUM: ["+TrackNumber", "+DisplayName"],
    models.Ref.ARTIST: ["+TypeEx", "+DisplayName"],
    models.Ref.DIRECTORY: ["+TypeEx", "+DisplayName"],
}

LOOKUP_QUERY = 'Type = "music" or Type = "audio"'  # TODO: check SearchCaps


def iterate(func, translate, limit):
    count = 0
    future = func(count, limit)
    while future:
        try:
            objs, more = future.get()
        except Exception:
            # some media servers will raise a DBusException when
            # iterating beyond the last item, which may happen with
            # Browse due to dLeyna not passing TotalMatches...
            if count:
                break
            else:
                raise
        if more:
            objs = list(objs)
            count += len(objs)
            future = func(count, limit)
        else:
            future = None
        for obj in objs:
            try:
                result = translate(obj)
            except ValueError as e:
                logger.debug("Skipping %s: %s", obj.get("URI"), e)
            else:
                yield result


class dLeynaLibraryProvider(backend.LibraryProvider):

    root_directory = models.Ref.directory(
        uri=uritools.uricompose(Extension.ext_name),
        name="Digital Media Servers",
    )

    def __init__(self, backend, config):
        super().__init__(backend)
        ext_config = config[Extension.ext_name]
        self.__upnp_browse_limit = ext_config["upnp_browse_limit"]
        self.__upnp_lookup_limit = ext_config["upnp_lookup_limit"]
        self.__upnp_search_limit = ext_config["upnp_search_limit"]

    def browse(self, uri):
        if uri == self.root_directory.uri:
            refs = sorted(self.__servers, key=operator.attrgetter("name"))
        else:
            refs = self.__browse(uri)
        return list(refs)

    def get_images(self, uris):
        # TODO: suggest as API improvement
        uris = frozenset(uris)
        # group uris by authority (media server)
        queries = collections.defaultdict(list)
        for uri in uris.difference([self.root_directory.uri]):
            parts = uritools.urisplit(uri)
            baseuri = parts.scheme + "://" + parts.authority
            queries[baseuri].append(parts.path)
        # start searching - blocks only when iterating over results
        results = []
        for baseuri, paths in queries.items():
            try:
                iterable = self.__images(baseuri, paths)
            except NotImplementedError as e:
                logger.warning("Not retrieving images for %s: %s", baseuri, e)
            else:
                results.append(iterable)
        # merge results
        result = {}
        for uri, images in itertools.chain.from_iterable(results):
            result[uri] = tuple(images)
        if self.root_directory.uri in uris:
            result[self.root_directory.uri] = tuple()
        return result

    def lookup(self, uri):
        if uri == self.root_directory.uri:
            tracks = []
        else:
            tracks = self.__lookup(uri)
        return list(tracks)

    def refresh(self, uri=None):
        logger.info("Refreshing dLeyna library")
        self.backend.client.rescan().get()

    def search(self, query=None, uris=None, exact=False):
        # sanitize uris - remove duplicates, replace root with server uris
        uris = set(uris or [self.root_directory.uri])
        if self.root_directory.uri in uris:
            uris.update(ref.uri for ref in self.__servers)
            uris.remove(self.root_directory.uri)
        # start searching - blocks only when iterating over results
        results = []
        for uri in uris:
            try:
                iterable = self.__search(uri, query, exact)
            except NotImplementedError as e:
                logger.warning("Not searching %s: %s", uri, e)
            else:
                results.append(iterable)
        if not results:
            return None
        # retrieve and merge search results - TODO: handle exceptions?
        result = collections.defaultdict(collections.OrderedDict)
        for model in itertools.chain.from_iterable(results):
            result[type(model)][model.uri] = model
        return models.SearchResult(
            albums=result[models.Album].values(),
            artists=result[models.Artist].values(),
            tracks=result[models.Track].values(),
        )

    def __browse(self, uri, filter=BROWSE_FILTER):
        client = self.backend.client
        obj = client.properties(uri, iface=client.MEDIA_OBJECT_IFACE).get()
        order = BROWSE_ORDER[translator.ref(obj).type]

        def browse(offset, limit):
            return client.browse(uri, offset, limit, filter, order)

        return iterate(browse, translator.ref, self.__upnp_browse_limit)

    def __images(self, baseuri, paths, filter=IMAGES_FILTER):
        client = self.backend.client
        server = client.server(baseuri).get()
        # fall back on properties if path search is not available/enabled
        if self.__upnp_lookup_limit == 1 or "Path" not in server["SearchCaps"]:
            futures = [client.properties(baseuri + path) for path in paths]
            return (translator.images(f.get()) for f in futures)
        # use path search for retrieving multiple results at once
        root = server["Path"]

        def images(offset, limit):
            slice = paths[offset : offset + limit if limit else None]
            query = " or ".join(f'Path = "{root}{p}"' for p in slice)
            return client.search(baseuri, query, 0, 0, filter).map(
                lambda res: (res[0], limit and offset + limit < len(paths))
            )

        return iterate(images, translator.images, self.__upnp_lookup_limit)

    def __lookup(self, uri, filter=LOOKUP_FILTER):
        client = self.backend.client
        obj = client.properties(uri).get()
        if translator.ref(obj).type == models.Ref.TRACK:
            objs = [obj]
        else:
            objs, _ = client.search(uri, LOOKUP_QUERY, filter=filter).get()
        return map(translator.track, objs)

    def __search(self, uri, query, exact, filter=SEARCH_FILTER):
        client = self.backend.client
        server = client.server(uri).get()
        if server["SearchCaps"]:
            q = translator.query(query or {}, exact, server["SearchCaps"])
        else:
            raise NotImplementedError("Search is not supported by this device")

        def search(offset, limit):
            return client.search(uri, q, offset, limit, filter)

        return iterate(search, translator.model, self.__upnp_search_limit)

    @property
    def __servers(self):
        for server in self.backend.client.servers().get():
            name = server.get("FriendlyName", server["DisplayName"])
            yield models.Ref.directory(name=name, uri=server["URI"])



import logging

from mopidy import backend

logger = logging.getLogger(__name__)


class dLeynaPlaybackProvider(backend.PlaybackProvider):
    def translate_uri(self, uri):
        # TODO: GetCompatibleResources w/protocol_info
        client = self.backend.client
        try:
            obj = client.properties(uri, client.MEDIA_ITEM_IFACE).get()
        except Exception as e:
            logger.error("Error translating %s: %s", uri, e)
        else:
            return obj["URLs"][0]



from unittest import mock

import pytest

from mopidy import models
from mopidy_dleyna.util import Future


@pytest.fixture
def container():
    return {
        "DisplayName": "Album #1",
        "Type": "container",
        "TypeEx": "container.album.musicAlbum",
        "URI": "dleyna://media/1",
    }


@pytest.fixture
def items():
    return [
        {
            "DisplayName": "Track #1",
            "Type": "music",
            "URI": "dleyna://media/11",
        },
        {
            "DisplayName": "Track #2",
            "Type": "audio",
            "URI": "dleyna://media/12",
        },
    ]


def test_lookup_root(backend):
    assert backend.library.lookup(backend.library.root_directory.uri) == []


def test_lookup_item(backend, items):
    with mock.patch.object(backend, "client") as m:
        m.properties.return_value = Future.fromvalue(items[0])
        assert backend.library.lookup(items[0]["URI"]) == [
            models.Track(name="Track #1", uri="dleyna://media/11")
        ]


def test_lookup_container(backend, container, items):
    with mock.patch.object(backend, "client") as m:
        m.properties.return_value = Future.fromvalue(container)
        m.search.return_value = Future.fromvalue([items, False])
        assert backend.library.lookup(container["URI"]) == [
            models.Track(name="Track #1", uri="dleyna://media/11"),
            models.Track(name="Track #2", uri="dleyna://media/12"),
        ]



from unittest import mock

import pytest

from mopidy_dleyna.util import Future


@pytest.fixture
def item():
    return {
        "DisplayName": "Track #1",
        "Type": "music",
        "URI": "dleyna://media/1",
        "URLs": ["http://example.com/1.mp3"],
    }


def test_translate_uri(backend, item):
    with mock.patch.object(backend, "client") as m:
        m.properties.return_value = Future.fromvalue(item)
        assert backend.playback.translate_uri(item["URI"]) == item["URLs"][0]


def test_translate_unknown_uri(backend):
    with mock.patch.object(backend, "client") as m:
        m.properties.return_value = Future.exception(LookupError("Not Found"))
        assert backend.playback.translate_uri("") is None



from unittest import mock

import pytest

from mopidy import models
from mopidy_dleyna.util import Future


@pytest.fixture
def server():
    return {
        "FriendlyName": "Media Server",
        "DisplayName": "Media",
        "SearchCaps": ["DisplayName", "Path"],
        "Path": "/com/intel/dLeynaServer/server/0",
        "URI": "dleyna://media",
    }


@pytest.fixture
def items():
    return [
        {
            "DisplayName": "Track #1",
            "Type": "music",
            "AlbumArtURL": "http:://example.com/1.jpg",
            "URI": "dleyna://media/1",
        },
        {"DisplayName": "Track #2", "Type": "audio", "URI": "dleyna://media/2"},
    ]


def test_images(backend, server, items):
    with mock.patch.object(backend, "client") as m:
        m.servers.return_value = Future.fromvalue([server])
        m.server.return_value = Future.fromvalue(server)
        m.search.return_value = Future.fromvalue([items, False])
        assert backend.library.get_images(item["URI"] for item in items) == {
            items[0]["URI"]: (models.Image(uri=items[0]["AlbumArtURL"]),),
            items[1]["URI"]: tuple(),
        }



from mopidy_dleyna import Extension


def test_get_default_config():
    config = Extension().get_default_config()
    assert "[" + Extension.ext_name + "]" in config
    assert "enabled = true" in config


def test_get_config_schema():
    schema = Extension().get_config_schema()
    assert "upnp_browse_limit" in schema
    assert "upnp_lookup_limit" in schema
    assert "upnp_search_limit" in schema
    assert "dbus_start_session" in schema



import logging
import time

import pykka

logger = logging.getLogger(__name__)


class Future(pykka.ThreadingFuture):

    Timeout = pykka.Timeout

    @classmethod
    def exception(cls, exc=None):
        future = cls()
        future.set_exception((type(exc), exc, None))
        return future

    @classmethod
    def fromdbus(cls, func, *args, **kwargs):
        method = getattr(func, "_method_name", "<unknown>")
        logger.debug("Calling D-Bus method %s%s", method, args)
        future = cls()
        start = time.time()

        def reply(*args):
            logger.debug("%s reply after %.3fs", method, time.time() - start)
            future.set(args[0] if len(args) == 1 else args)

        def error(e):
            logger.debug("%s error after %.3fs", method, time.time() - start)
            future.set_exception(exc_info=(type(e), e, None))

        func(*args, reply_handler=reply, error_handler=error, **kwargs)
        return future

    @classmethod
    def fromvalue(cls, value):
        future = cls()
        future.set(value)
        return future



import errno
import logging
import os
import os.path
import re
import signal
import stat
import subprocess

import pykka

from mopidy import backend, exceptions

from . import Extension
from .client import dLeynaClient
from .library import dLeynaLibraryProvider
from .playback import dLeynaPlaybackProvider

DBUS_SESSION_BUS_RE = re.compile(
    r"""
^(?:DBUS_SESSION_BUS_ADDRESS=)?(.*)\n
^(?:DBUS_SESSION_BUS_PID=)?(\d+)$
""",
    re.MULTILINE | re.VERBOSE,
)

logger = logging.getLogger(__name__)


class dLeynaBackend(pykka.ThreadingActor, backend.Backend):

    uri_schemes = [Extension.ext_name]

    __dbus_pid = None

    def __init__(self, config, audio):
        super().__init__()
        try:
            if self.__have_session_bus():
                self.client = dLeynaClient()
            else:
                command = config[Extension.ext_name]["dbus_start_session"]
                address, self.__dbus_pid = self.__start_session_bus(command)
                self.client = dLeynaClient(address)
        except Exception as e:
            logger.error("Error starting %s: %s", Extension.dist_name, e)
            # TODO: clean way to bail out late?
            raise exceptions.ExtensionError("Error starting dLeyna client")
        self.library = dLeynaLibraryProvider(self, config)
        self.playback = dLeynaPlaybackProvider(audio, self)

    def on_stop(self):
        if self.__dbus_pid is not None:
            self.__stop_session_bus(self.__dbus_pid)

    def __have_session_bus(self):
        if "DBUS_SESSION_BUS_ADDRESS" in os.environ:
            return True
        if "XDG_RUNTIME_DIR" not in os.environ:
            return False
        try:
            st = os.stat(os.path.expandvars("$XDG_RUNTIME_DIR/bus"))
        except OSError:
            return False
        else:
            return stat.S_ISSOCK(st.st_mode) and st.st_uid == os.geteuid()

    def __start_session_bus(self, command):
        logger.info("Starting %s D-Bus daemon", Extension.dist_name)
        out = subprocess.check_output(command.split(), universal_newlines=True)
        logger.debug('Running "%s" returned:\n%s', command, out)
        match = DBUS_SESSION_BUS_RE.search(out)
        if not match:
            raise ValueError(f"{command} returned invalid output: {out}")
        else:
            return str(match.group(1)), int(match.group(2))

    def __stop_session_bus(self, pid):
        logger.error("Stopping %s D-Bus daemon (%d)", Extension.dist_name, pid)
        try:
            os.kill(pid, signal.SIGTERM)
        except OSError as e:
            if e.errno != errno.ESRCH:
                raise
        logger.debug("Stopped %s D-Bus daemon", Extension.dist_name)



import pytest

from mopidy.models import Album, Artist, Ref, Track
from mopidy_dleyna import translator

BASEURI = "dleyna://foo"
BASEPATH = "/com/intel/dLeynaServer/server/0"

ALBUM_TYPE = "container.album.musicAlbum"
ARTIST_TYPE = "container.person.musicArtist"
PLAYLIST_TYPE = "container.playlistContainer"
AUDIO_BOOK_TYPE = "item.audioItem.audioBook"
AUDIO_BROADCAST_TYPE = "item.audioItem.audioBroadcast"


def test_album_ref():
    assert translator.ref(
        {
            "DisplayName": "Foo",
            "URI": BASEURI + "/foo",
            "Type": "container",
            "TypeEx": ALBUM_TYPE,
        }
    ) == Ref.album(uri=BASEURI + "/foo", name="Foo")


def test_artist_ref():
    assert translator.ref(
        {
            "DisplayName": "Foo",
            "URI": BASEURI + "/foo",
            "Type": "container",
            "TypeEx": ARTIST_TYPE,
        }
    ) == Ref.artist(uri=BASEURI + "/foo", name="Foo")


def test_music_ref():
    assert translator.ref(
        {"DisplayName": "Foo", "URI": BASEURI + "/foo", "Type": "music"}
    ) == Ref.track(uri=BASEURI + "/foo", name="Foo")


def test_audio_book_ref():
    assert translator.ref(
        {
            "DisplayName": "Foo",
            "URI": BASEURI + "/foo",
            "Type": "audio",
            "TypeEx": AUDIO_BOOK_TYPE,
        }
    ) == Ref.track(uri=BASEURI + "/foo", name="Foo")


def test_audio_broadcast_ref():
    assert translator.ref(
        {
            "DisplayName": "Foo",
            "URI": BASEURI + "/foo",
            "Type": "audio",
            "TypeEx": AUDIO_BROADCAST_TYPE,
        }
    ) == Ref.track(uri=BASEURI + "/foo", name="Foo")


def test_container_ref():
    assert translator.ref(
        {"DisplayName": "Foo", "URI": BASEURI + "/foo", "Type": "container"}
    ) == Ref.directory(uri=BASEURI + "/foo", name="Foo")


def test_video_ref():
    with pytest.raises(ValueError):
        translator.ref(
            {"DisplayName": "Foo", "URI": BASEURI + "/foo", "Type": "video"}
        )


def test_album():
    assert translator.model(
        {
            "DisplayName": "Foo",
            "URI": BASEURI + "/foo",
            "Type": "container",
            "TypeEx": ALBUM_TYPE,
        }
    ) == Album(uri=BASEURI + "/foo", name="Foo")


def test_artist():
    assert translator.model(
        {
            "DisplayName": "Foo",
            "URI": BASEURI + "/foo",
            "Type": "container",
            "TypeEx": ARTIST_TYPE,
        }
    ) == Artist(uri=BASEURI + "/foo", name="Foo")


def test_playlist():
    assert translator.ref(
        {
            "DisplayName": "Foo",
            "URI": BASEURI + "/foo",
            "Type": "container",
            "TypeEx": PLAYLIST_TYPE,
        }
    ) == Ref.directory(uri=BASEURI + "/foo", name="Foo")


def test_track():
    assert translator.model(
        {
            "DisplayName": "Foo",
            "Album": "Bar",
            "URI": BASEURI + "/foo",
            "Type": "music",
        }
    ) == Track(uri=BASEURI + "/foo", name="Foo", album=Album(name="Bar"))


def test_audio_book():
    assert translator.model(
        {
            "DisplayName": "Foo",
            "URI": BASEURI + "/foo",
            "Type": "audio",
            "TypeEx": AUDIO_BOOK_TYPE,
        }
    ) == Track(uri=BASEURI + "/foo", name="Foo")


def test_audio_broadcast():
    assert translator.model(
        {
            "DisplayName": "Foo",
            "URI": BASEURI + "/foo",
            "Type": "audio",
            "TypeEx": AUDIO_BROADCAST_TYPE,
        }
    ) == Track(uri=BASEURI + "/foo", name="Foo")


def test_video():
    with pytest.raises(ValueError):
        translator.model(
            {"DisplayName": "Foo", "URI": BASEURI + "/foo", "Type": "video"}
        )



from unittest import mock

import pytest

from mopidy import models
from mopidy_dleyna.util import Future


@pytest.fixture
def server():
    return {
        "FriendlyName": "Media Server",
        "DisplayName": "Media",
        "SearchCaps": ["DisplayName"],
        "URI": "dleyna://media",
    }


@pytest.fixture
def result():
    return [
        {
            "DisplayName": "Album #1",
            "Type": "container",
            "TypeEx": "container.album.musicAlbum",
            "URI": "dleyna://media/1",
        },
        {
            "DisplayName": "Track #1",
            "Type": "music",
            "URI": "dleyna://media/11",
        },
        {
            "DisplayName": "Track #2",
            "Type": "audio",
            "URI": "dleyna://media/12",
        },
    ]


def test_search(backend, server, result):
    with mock.patch.object(backend, "client") as m:
        m.servers.return_value = Future.fromvalue([server])
        m.server.return_value = Future.fromvalue(server)
        m.search.side_effect = [
            Future.fromvalue([result[0:2], True]),
            Future.fromvalue([result[2:3], False]),
        ]
        # valid search
        assert backend.library.search({"any": ["foo"]}) == models.SearchResult(
            albums=[models.Album(name="Album #1", uri="dleyna://media/1")],
            tracks=[
                models.Track(name="Track #1", uri="dleyna://media/11"),
                models.Track(name="Track #2", uri="dleyna://media/12"),
            ],
        )
        # unsupported search field yields no result
        assert backend.library.search({"composer": ["foo"]}) is None
        # search field not supported by device yields no result
        assert backend.library.search({"genre": ["foo"]}) is None



import pathlib

import pkg_resources

from mopidy import config, exceptions, ext

__version__ = pkg_resources.get_distribution("Mopidy-dLeyna").version


class Extension(ext.Extension):

    dist_name = "Mopidy-dLeyna"
    ext_name = "dleyna"
    version = __version__

    def get_default_config(self):
        return config.read(pathlib.Path(__file__).parent / "ext.conf")

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["upnp_browse_limit"] = config.Integer(minimum=0)
        schema["upnp_lookup_limit"] = config.Integer(minimum=0)
        schema["upnp_search_limit"] = config.Integer(minimum=0)
        schema["dbus_start_session"] = config.String()
        return schema

    def setup(self, registry):
        from .backend import dLeynaBackend

        registry.add("backend", dLeynaBackend)

    def validate_environment(self):
        try:
            import dbus  # noqa
        except ImportError:
            raise exceptions.ExtensionError("Cannot import dbus")



from mopidy import models

_QUERY = {
    "any": lambda caps: (
        " or ".join(
            s + ' {0} "{1}"'
            for s in caps & {"DisplayName", "Album", "Artist", "Genre"}
        )
    ),
    "album": lambda caps: ('Album {0} "{1}"' if "Album" in caps else None),
    "albumartist": lambda caps: (
        'Artist {0} "{1}" and TypeEx = "container.album.musicAlbum"'
        if "Artist" in caps and "TypeEx" in caps
        else None
    ),
    "artist": lambda caps: ('Artist {0} "{1}"' if "Artist" in caps else None),
    "date": lambda caps: (
        'Date = "{1}"' if "Date" in caps else None  # TODO: inexact?
    ),
    "genre": lambda caps: ('Genre {0} "{1}"' if "Genre" in caps else None),
    "track_name": lambda caps: (
        'DisplayName {0} "{1}" and Type = "music"'
        if "DisplayName" in caps and "Type" in caps
        else None
    ),
    "track_no": lambda caps: (
        'TrackNumber = "{1}"' if "TrackNumber" in caps else None
    ),
}

_REFS = {
    "audio": models.Ref.track,
    "container": models.Ref.directory,
    "container.album": models.Ref.directory,
    "container.album.musicAlbum": models.Ref.album,
    "container.genre.musicGenre": models.Ref.directory,
    "container.person.musicArtist": models.Ref.artist,
    "container.playlistContainer": models.Ref.directory,
    "container.storageFolder": models.Ref.directory,
    "item.audioItem.audioBook": models.Ref.track,
    "item.audioItem.audioBroadcast": models.Ref.track,
    "music": models.Ref.track,
}


def _album(obj):
    try:
        name = obj["Album"]
    except KeyError:
        return None
    else:
        return models.Album(name=name, uri=None)


def _artists(obj):
    return (models.Artist(name=name) for name in obj.get("Artists", []))


def _quote(s):
    return str(s).replace("\\", "\\\\").replace('"', '\\"')


def ref(obj):
    type = obj.get("TypeEx", obj["Type"])
    try:
        translate = _REFS[type]
    except KeyError:
        raise ValueError('Object type "%s" not supported' % type)
    else:
        return translate(
            name=obj.get("DisplayName", obj["URI"]), uri=obj["URI"]
        )


def album(obj):
    return models.Album(
        uri=obj["URI"],
        name=obj["DisplayName"],
        artists=list(_artists(obj)),
        num_tracks=obj.get("ItemCount", obj.get("ChildCount")),
    )


def artist(obj):
    return models.Artist(name=obj["DisplayName"], uri=obj["URI"])


def track(obj):
    return models.Track(
        uri=obj["URI"],
        name=obj["DisplayName"],
        artists=list(_artists(obj)),
        album=_album(obj),
        genre=obj.get("Genre"),
        track_no=obj.get("TrackNumber"),
        date=obj.get("Date"),
        length=obj.get("Duration", 0) * 1000 or None,
        bitrate=obj.get("Bitrate", 0) * 8 or None,
    )


def model(obj):
    type = obj.get("TypeEx", obj["Type"])
    ref = _REFS.get(type)
    if ref == models.Ref.track:
        return track(obj)
    elif ref == models.Ref.album:
        return album(obj)
    elif ref == models.Ref.artist:
        return artist(obj)
    else:
        raise ValueError('Object type "%s" not supported' % type)


def images(obj):
    if "AlbumArtURL" in obj:
        return obj["URI"], [models.Image(uri=str(obj["AlbumArtURL"]))]
    else:
        return obj["URI"], []


def query(query, exact, searchcaps):
    op = "=" if exact else "contains"
    terms = []
    for key, values in query.items():
        try:
            translate = _QUERY[key]
        except KeyError:
            raise NotImplementedError('Keyword "%s" not supported' % key)
        else:
            fmt = translate(frozenset(searchcaps))
        if fmt:
            terms.extend(fmt.format(op, _quote(value)) for value in values)
        else:
            raise NotImplementedError('Keyword "%s" not searchable' % key)
    return ("(%s)" % ") and (".join(terms)) or "*"



from collections.abc import Mapping
import logging
import threading
import time

import dbus
import uritools

from . import Extension, util

SERVER_BUS_NAME = "com.intel.dleyna-server"

SERVER_ROOT_PATH = "/com/intel/dLeynaServer"

SERVER_MANAGER_IFACE = "com.intel.dLeynaServer.Manager"

logger = logging.getLogger(__name__)


def urifilter(fields):
    if "URI" in fields:
        objfilter = fields[:]
        objfilter.remove("URI")
        objfilter.append("Path")
        objfilter.append("RefPath")
        return objfilter
    else:
        return fields


def urimapper(baseuri, prefix="/com/intel/dLeynaServer/server/"):
    default_index = len(prefix)

    def mapper(obj, index=default_index):
        objpath = obj.get("RefPath", obj["Path"])
        assert objpath.startswith(prefix)
        _, sep, relpath = objpath[index:].partition("/")
        obj["URI"] = baseuri + sep + relpath
        return obj

    return mapper


class Servers(Mapping):
    def __init__(self, bus):
        self.__bus = bus
        self.__lock = threading.RLock()
        self.__servers = {}

        bus.add_signal_receiver(
            self.__found_server, "FoundServer", bus_name=SERVER_BUS_NAME
        )
        bus.add_signal_receiver(
            self.__lost_server, "LostServer", bus_name=SERVER_BUS_NAME
        )
        self.__get_servers()

    def __getitem__(self, udn):
        key = udn.lower()
        with self.__lock:
            return self.__servers[key]

    def __iter__(self):
        with self.__lock:
            return iter(list(self.__servers))

    def __len__(self):
        with self.__lock:
            return len(self.__servers)

    def __add_server(self, obj):
        udn = obj["UDN"]
        obj["URI"] = uritools.uricompose(Extension.ext_name, udn)
        obj["DisplayName"] = obj.get("DisplayName", obj["URI"])
        key = udn.lower()
        if key not in self:
            self.__log_server_action("Found", obj)
        with self.__lock:
            self.__servers[key] = obj

    def __remove_server(self, obj):
        key = obj["UDN"].lower()
        with self.__lock:
            del self.__servers[key]
        self.__log_server_action("Lost", obj)

    def __found_server(self, path):
        def error_handler(e):
            logger.warning("Cannot access media server %s: %s", path, e)

        self.__bus.get_object(SERVER_BUS_NAME, path).GetAll(
            "",  # all interfaces
            dbus_interface=dbus.PROPERTIES_IFACE,
            reply_handler=self.__add_server,
            error_handler=error_handler,
        )

    def __lost_server(self, path):
        with self.__lock:
            servers = list(self.__servers.values())
        for obj in servers:
            if obj["Path"] == path:
                return self.__remove_server(obj)
        logger.info("Lost digital media server %s", path)

    def __get_servers(self):
        def reply_handler(paths):
            for path in paths:
                self.__found_server(path)

        def error_handler(e):
            logger.error("Cannot retrieve digital media servers: %s", e)

        self.__bus.get_object(SERVER_BUS_NAME, SERVER_ROOT_PATH).GetServers(
            dbus_interface=SERVER_MANAGER_IFACE,
            reply_handler=reply_handler,
            error_handler=error_handler,
        )

    @classmethod
    def __log_server_action(cls, action, obj):
        logger.info(
            "%s digital media server %s [%s]",
            action,
            obj["FriendlyName"],
            obj["UDN"],
        )


class dLeynaClient:

    MEDIA_CONTAINER_IFACE = "org.gnome.UPnP.MediaContainer2"

    MEDIA_DEVICE_IFACE = "com.intel.dLeynaServer.MediaDevice"

    MEDIA_ITEM_IFACE = "org.gnome.UPnP.MediaItem2"

    MEDIA_OBJECT_IFACE = "org.gnome.UPnP.MediaObject2"

    def __init__(self, address=None, mainloop=None):
        if address:
            self.__bus = dbus.bus.BusConnection(address, mainloop=mainloop)
        else:
            self.__bus = dbus.SessionBus(mainloop=mainloop)
        self.__servers = Servers(self.__bus)

    def browse(self, uri, offset=0, limit=0, filter=None, order=None):
        if filter is None:
            filter = ["*"]
        if order is None:
            order = []

        baseuri, objpath = self.__parseuri(uri)
        future = util.Future.fromdbus(
            self.__bus.get_object(SERVER_BUS_NAME, objpath).ListChildrenEx,
            dbus.UInt32(offset),
            dbus.UInt32(limit),
            urifilter(filter),
            ",".join(self.__sortorder(uri, order)),
            dbus_interface=self.MEDIA_CONTAINER_IFACE,
        )

        def mapper(res):
            if baseuri and (filter == ["*"] or "URI" in filter):
                objs = map(urimapper(baseuri), res)
            else:
                objs = res
            # dleyna does not pass TotalMatches from Browse action;
            # also note that a server may choose to return less than
            # `limit` items, so assume `more` until nothing found
            more = len(res) != 0
            return objs, more

        return future.map(mapper)

    def properties(self, uri, iface=None):
        baseuri, objpath = self.__parseuri(uri)
        future = util.Future.fromdbus(
            self.__bus.get_object(SERVER_BUS_NAME, objpath).GetAll,
            iface or "",
            dbus_interface=dbus.PROPERTIES_IFACE,
        )
        if baseuri and (not iface or iface == self.MEDIA_OBJECT_IFACE):
            return future.map(urimapper(baseuri))
        else:
            return future

    def rescan(self):
        return util.Future.fromdbus(
            self.__bus.get_object(SERVER_BUS_NAME, SERVER_ROOT_PATH).Rescan,
            dbus_interface=SERVER_MANAGER_IFACE,
        )

    def search(self, uri, query, offset=0, limit=0, filter=None, order=None):
        if filter is None:
            filter = ["*"]
        if order is None:
            order = []

        baseuri, objpath = self.__parseuri(uri)
        future = util.Future.fromdbus(
            self.__bus.get_object(SERVER_BUS_NAME, objpath).SearchObjectsEx,
            query,
            dbus.UInt32(offset),
            dbus.UInt32(limit),
            urifilter(filter),
            ",".join(self.__sortorder(uri, order)),
            dbus_interface=self.MEDIA_CONTAINER_IFACE,
        )

        def mapper(res):
            items, total = res
            if baseuri and (filter == ["*"] or "URI" in filter):
                objs = map(urimapper(baseuri), items)
            else:
                objs = items
            more = offset + len(items) < total
            return objs, more

        return future.map(mapper)

    def server(self, uri):
        # return future for consistency/future extensions
        return util.Future.fromvalue(self.__server(uri))

    def servers(self):
        # return future for consistency/future extensions
        return util.Future.fromvalue(self.__servers.values())

    def __parseuri(self, uri):
        try:
            server = self.__server(uri)
        except ValueError:
            return None, uri
        else:
            return server["URI"], server["Path"] + uritools.urisplit(uri).path

    def __sortorder(self, uri, order):
        try:
            server = self.__server(uri)
        except ValueError:
            sortcaps = frozenset("*")
        else:
            sortcaps = frozenset(server.get("SortCaps", []))
        if "*" in sortcaps:
            return order
        else:
            return list(filter(lambda f: f[1:] in sortcaps, order))

    def __server(self, uri):
        udn = uritools.urisplit(uri).gethost()
        if not udn:
            raise ValueError("Invalid URI %s" % uri)
        try:
            server = self.__servers[udn]
        except KeyError:
            raise LookupError("Unknown media server UDN %s" % udn)
        else:
            return server


if __name__ == "__main__":  # pragma: no cover
    import argparse
    import json
    import sys

    import dbus.mainloop.glib
    import gobject

    parser = argparse.ArgumentParser()
    parser.add_argument("uri", metavar="PATH | URI", nargs="?")
    parser.add_argument("-b", "--browse", action="store_true")
    parser.add_argument("-f", "--filter", default=["*"], nargs="*")
    parser.add_argument("-i", "--iface")
    parser.add_argument("-m", "--offset", default=0, type=int)
    parser.add_argument("-n", "--limit", default=0, type=int)
    parser.add_argument("-o", "--order", default=[], nargs="*")
    parser.add_argument("-q", "--query")
    parser.add_argument("-t", "--timeout", default=1.0, type=float)
    parser.add_argument("-v", "--verbose", action="store_true")
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.WARN)

    client = dLeynaClient(mainloop=dbus.mainloop.glib.DBusGMainLoop())

    start = time.time()
    while time.time() < start + args.timeout:
        if args.uri:
            try:
                client.server(args.uri).get()
            except LookupError:
                pass
            except ValueError:
                break  # D-BUS path given
            else:
                break
        gobject.MainLoop().get_context().iteration(False)

    kwargs = {
        "offset": args.offset,
        "limit": args.limit,
        "filter": args.filter,
        "order": args.order,
    }

    if not args.uri:
        future = client.servers()
    elif args.browse:
        future = client.browse(args.uri, **kwargs)
    elif args.query:
        future = client.search(args.uri, args.query, **kwargs)
    else:
        future = client.properties(args.uri, iface=args.iface)

    while True:
        try:
            future.get(timeout=0)
        except util.Future.Timeout:
            gobject.MainLoop().get_context().iteration(True)
        else:
            break

    json.dump(future.get(), sys.stdout, default=vars, indent=2)
    sys.stdout.write("\n")



from unittest import mock

import pytest

from mopidy.models import Ref
from mopidy_dleyna.util import Future


@pytest.fixture
def servers():
    return [
        {
            "FriendlyName": "Media Server #1",
            "DisplayName": "Media1",
            "URI": "dleyna://media1",
        },
        {
            "FriendlyName": "Media Server #2",
            "DisplayName": "Media2",
            "URI": "dleyna://media2",
        },
    ]


@pytest.fixture
def container():
    return {"DisplayName": "Root", "Type": "container", "URI": "dleyna://media"}


@pytest.fixture
def items():
    return [
        {"DisplayName": "Track #1", "Type": "music", "URI": "dleyna://media/1"},
        {"DisplayName": "Track #2", "Type": "audio", "URI": "dleyna://media/2"},
        {"DisplayName": "Track #3", "Type": "audio", "URI": "dleyna://media/3"},
    ]


def test_browse_root(backend, servers):
    with mock.patch.object(backend, "client") as m:
        m.servers.return_value = Future.fromvalue(servers)
        assert backend.library.browse(backend.library.root_directory.uri) == [
            Ref.directory(name="Media Server #1", uri="dleyna://media1"),
            Ref.directory(name="Media Server #2", uri="dleyna://media2"),
        ]


def test_browse_items(backend, container, items):
    # FIXME: how to patch multiple object methods...
    with mock.patch.object(backend, "client") as m:
        m.properties.return_value = Future.fromvalue(container)
        m.browse.side_effect = [
            Future.fromvalue([items[0:2], True]),
            Future.fromvalue([items[2:3], True]),
            Future.fromvalue([[], False]),
        ]
        assert backend.library.browse(container["URI"]) == [
            Ref.track(name="Track #1", uri="dleyna://media/1"),
            Ref.track(name="Track #2", uri="dleyna://media/2"),
            Ref.track(name="Track #3", uri="dleyna://media/3"),
        ]
